import{p as e,f as o}from"./flowDb-956e92f1-10044f05.js";import{f as t,g as a}from"./styles-c10674c1-7b3680cf.js";import{u as i}from"./index-3a23e736.js";import"./graph-4ede11ff.js";import"./layout-dd48f7f4.js";import"./index-3862675e-65549d37.js";import"./clone-edb7f119.js";import"./edges-e0da2a9e-5a8b0749.js";import"./createText-2e5e7dd3-87c4ece5.js";import"./line-1569ad2c.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-dfa68278.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
